;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "mcp" "20260508.1208"
  "Model Context Protocol."
  '((emacs   "30.1")
    (jsonrpc "1.0.25"))
  :url "https://github.com/lizqwerscott/mcp.el"
  :commit "f10768e16f94f65527a0ea657ec91ab2eeaf244d"
  :revdesc "f10768e16f94"
  :keywords '("tools")
  :authors '(("lizqwer scott" . "lizqwerscott@gmail.com"))
  :maintainers '(("lizqwer scott" . "lizqwerscott@gmail.com")))
